from enum import Enum


class EnvironmentVariable(Enum):
    ENVIRONMENT: str = "ENVIRONMENT"
    SENTRY_URL: str = "SENTRY_URL"
    MONGO_DB_CONNECTION_STRING: str = "MONGODB_CONNECTION_STRING"
    DATABASE_NAME: str = "DATABASE_NAME"
    DISCORD_BOT_TOKEN: str = "DISCORD_BOT_TOKEN"
    APPLICATION_NAME: str = "APPLICATION_NAME"
    WISE_PRIMARY_ACCOUNT_API_KEY: str = "WISE_PRIMARY_ACCOUNT_API_KEY"
    WISE_SECONDARY_ACCOUNT_API_KEY: str = "WISE_SECONDARY_ACCOUNT_API_KEY"
    WISE_SANDBOX_ACCOUNT_API_KEY: str = "WISE_SANDBOX_ACCOUNT_API_KEY"
